/*
 * Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
 *
 * NVIDIA CORPORATION and its licensors retain all intellectual property
 * and proprietary rights in and to this software, related documentation
 * and any modifications thereto.  Any use, reproduction, disclosure or
 * distribution of this software and related documentation without an express
 * license agreement from NVIDIA CORPORATION is strictly prohibited.
 */

#pragma once

#include_next <stdlib.h>

#ifdef __cplusplus
extern "C"
#endif
void* __host_malloc(size_t);

#ifdef __cplusplus
extern "C"
#endif
void __host_free(void*);
