#pragma once

/* #undef LINALG_ENABLE_ATOMIC_REF */
/* #undef LINALG_ENABLE_BLAS */
#define LINALG_ENABLE_CONCEPTS
/* #undef LINALG_ENABLE_KOKKOS */
/* #undef LINALG_ENABLE_KOKKOS_DEFAULT */
/* #undef LINALG_ENABLE_TBB */
#define LINALG_FIX_CONJUGATED_FOR_NONCOMPLEX
#define LINALG_FIX_RANK_UPDATES
#define LINALG_FIX_TRANSPOSED_FOR_PADDED_LAYOUTS
