/* Copyright 2018, NVIDIA CORPORATION.  All rights reserved. */

/* <bits/link.h> uses __int128_t without first checking _GLIBCXX_USE_INT128,
   causing compilation errors with PGI, which does not support 128-bit
   integers.  To work around this, temporarily define __int128_t to be a
   128-bit type.  Any attempt to access the data members of type __int128_t in
   user code should fail, but they should be accessible through system library
   functions. */

#define __int128_t struct { long long __a, __b; }
#include_next <bits/link.h>
#undef __int128_t
