// Copyright 2020, NVIDIA CORPORATION. All rights reserved.
//
// NVIDIA CORPORATION and its licensors retain all intellectual property
// and proprietary rights in and to this software, related documentation
// and any modifications thereto.  Any use, reproduction, disclosure or
// distribution of this software and related documentation without an express
// license agreement from NVIDIA CORPORATION is strictly prohibited.

#pragma once

#include <nvhpc/pstl_config.hpp>

// If using the NVIDIA parallel algorithm implementation, this file is a no-op.
// If using the GCC implementation, include the real file.

#if _NVHPC_STDPAR_USE_GCC

#include_next <pstl/glue_memory_defs.h>

#endif
