/*
 *     Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
 *
 * NVIDIA CORPORATION and its licensors retain all intellectual property
 * and proprietary rights in and to this software, related documentation
 * and any modifications thereto.  Any use, reproduction, disclosure or
 * distribution of this software and related documentation without an express
 * license agreement from NVIDIA CORPORATION is strictly prohibited.
 *
 */

#ifndef _NVHPC_ACC_DEVICE_TYPE
#define _NVHPC_ACC_DEVICE_TYPE

#define nvhpc_acc_device_none 0
#define nvhpc_acc_device_default 1
#define nvhpc_acc_device_host 2
#define nvhpc_acc_device_not_host 3
#define nvhpc_acc_device_nvidia 4
#define nvhpc_acc_device_current 10

// These are device numbers that cannot be reused as they have historical
// and/or existing meaning.
#define nvhpc_acc_device_RESERVED5 5
#define nvhpc_acc_device_RESERVED6 6
#define nvhpc_acc_device_RESERVED7 7
#define nvhpc_acc_device_RESERVED8 8
#define nvhpc_acc_device_RESERVED9 9
#define nvhpc_acc_device_RESERVED1000 1000

#endif  // _NVHPC_ACC_DEVICE_TYPE