/* Copyright 2017-2026, NVIDIA CORPORATION. All rights reserved. */

#define __PGI_CPLUS_PREINCLUDE__ 1

#define __builtin_va_start(ap, param) (__builtin_pgi_va_start(ap, param))
#define __builtin_c23_va_start(ap, param) (__builtin_pgi_va_start(ap, param))
#define __builtin_va_arg(ap, type) (*(type*)__builtin_pgi_va_arg(ap, (type*)0))
#define __builtin_va_copy(dst, src)   ((dst)[0] = (src)[0])
#define __builtin_va_end(ap)          ((void)0)
