/* 
 *     Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
 *
 * NVIDIA CORPORATION and its licensors retain all intellectual property
 * and proprietary rights in and to this software, related documentation
 * and any modifications thereto.  Any use, reproduction, disclosure or
 * distribution of this software and related documentation without an express
 * license agreement from NVIDIA CORPORATION is strictly prohibited.
 *
 */

/*
 * iso_fortran_binding.h
 */

#ifndef _ISO_FORTRAN_BINDING_H
#define _ISO_FORTRAN_BINDING_H

#if !defined(__F18)
#define __OVERRIDE_CFI_MAX_RANK
#define CFI_MAX_RANK 7
#endif

#include <ISO_Fortran_binding_1.h>

#endif
