/* 
 *     Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
 *
 * NVIDIA CORPORATION and its licensors retain all intellectual property
 * and proprietary rights in and to this software, related documentation
 * and any modifications thereto.  Any use, reproduction, disclosure or
 * distribution of this software and related documentation without an express
 * license agreement from NVIDIA CORPORATION is strictly prohibited.
 *
 */

#if __pgnu_vsn >= 50200 && __pgnu_vsn <= 50300

/* The GCC 5.2 version of <bits/alloc_traits.h> causes linker errors about
   undefined references to std::string::_M_create when compiling with pgc++.
   Work around this by providing the GCC 5.3.1 version of <bits/alloc_traits.h>
   instead. */
#include <bits/alloc_traits_531.h>

#else
#include_next <bits/alloc_traits.h>
#endif

