//@HEADER
// ************************************************************************
//
//                        Kokkos v. 4.0
//       Copyright (2022) National Technology & Engineering
//               Solutions of Sandia, LLC (NTESS).
//
// Under the terms of Contract DE-NA0003525 with NTESS,
// the U.S. Government retains certain rights in this software.
//
// Part of Kokkos, under the Apache License v2.0 with LLVM Exceptions.
// See https://kokkos.org/LICENSE for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
// ************************************************************************
//@HEADER

#ifndef __LINALG___P1673_BITS_CONJUGATE_TRANSPOSED_HPP_
#define __LINALG___P1673_BITS_CONJUGATE_TRANSPOSED_HPP_

#include "conjugated.hpp"
#include "transposed.hpp"

namespace _NVHPC_MDSPAN_IMPL_STANDARD_NAMESPACE {
inline namespace __p1673_version_0 {
namespace linalg {

template<class ElementType, class Extents, class Layout, class Accessor>
auto conjugate_transposed(mdspan<ElementType, Extents, Layout, Accessor> a)
{
  return conjugated(transposed(a));
}

} // end namespace linalg
} // end inline namespace __p1673_version_0
} // end namespace _NVHPC_MDSPAN_IMPL_STANDARD_NAMESPACE

#endif //__LINALG___P1673_BITS_CONJUGATE_TRANSPOSED_HPP_
