//@HEADER
// ************************************************************************
//
//                        Kokkos v. 4.0
//       Copyright (2022) National Technology & Engineering
//               Solutions of Sandia, LLC (NTESS).
//
// Under the terms of Contract DE-NA0003525 with NTESS,
// the U.S. Government retains certain rights in this software.
//
// Part of Kokkos, under the Apache License v2.0 with LLVM Exceptions.
// See https://kokkos.org/LICENSE for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//@HEADER

#ifndef MDARRAY_HPP_
#define MDARRAY_HPP_

#ifndef _NVHPC_MDSPAN_IMPL_STANDARD_NAMESPACE
  #define _NVHPC_MDSPAN_IMPL_STANDARD_NAMESPACE Kokkos
#endif

#ifndef _NVHPC_MDSPAN_IMPL_PROPOSED_NAMESPACE
  #define _NVHPC_MDSPAN_IMPL_PROPOSED_NAMESPACE Experimental
#endif

#include "../__linalg/mdspan.hpp"
#include "__p1684_bits/mdarray.hpp"

#endif // MDARRAY_HPP_
